/* 
 * --------------------------------------------------------------------------
 * Common library functions for the haserl suite
 *
 * $Id: common.c,v 1.10 2005/10/27 14:09:39 nangel Exp $ 
 * Copyright (c) 2005 Nathan Angelacos (nangel@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * -------------------------------------------------------------------------
 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#if HAVE_SIGNAL_H
#include <signal.h>
#endif

#include "common.h"

/*
 * split a string into an argv[] array, and return the number of elements.
 * Warning:  Overwrites instr with nulls (to make ASCIIZ strings) and mallocs
 * space for the argv array.  The argv array will point to the offsets in
 * instr where the elements occur.  The calling function must free the argv
 * array, and should do so before freeing instr memory.
 *
 * If comment points to a non-null string, then any character in the string
 * will mark the beginning of a comment until the end-of-line:
 *
 * comment="#;"
 * foo bar     # This is a comment
 * foo baz     ; This is a comment as well
 * 
 * Example of use:
 *
 * int argc, count; char **argv; char string[2000];
 *
 * strcpy (string, "This\\ string will be \"separated into\" '6' elements.", "");
 * argc = argc_argv (string, &argv); 
 * for (count = 0; count < argc; count++) {
 * 	printf ("%03d: %s\n", count, argv[count]); 
 * 	} 
 * free (argv);
 *
 */

int
argc_argv (char *instr, char **argv[], char *commentstr)
{
  char quote = '\0';
  int arg_count = 0;
  enum state_t
  {
    WHITESPACE, WORDSPACE, TOKENSTART
  } state = WHITESPACE;
  char **argv_array = NULL;
  int argc_slots = 0;
  size_t len, pos;
  
  len= strlen(instr);
  pos=0;

  while (pos < len)
    {
    // printf ("%3d of %3d: %s\n", pos, len, instr);

      /* Comments are really, really special */

     if ((state == WHITESPACE) && (strchr(commentstr, *instr))) {
		     while ((*instr != '\n') && (*instr != '\0' )) { instr++; pos++; }
		}
	    
      switch (*instr)
	{

	  /* quoting */
	case '"':
	case '\'':
	  /* Begin quoting */
	  if (state == WHITESPACE )
	    {
	      quote = *instr;
	      state = TOKENSTART;
	      if (*(instr+1) == quote ) { /* special case for NULL quote */
	      	quote= '\0';
		*instr = '\0';
	      	} else {
	      instr++;
	      pos++;
		}
	    }
	  else /* WORDSPACE, so quotes end or quotes within quotes */
	    {
	      /* Is it the same kind of quote? */
	      if (*instr == quote)
		{
		  quote = '\0';
		  *instr = '\0';
		  state = WHITESPACE;
		}
	    }
	  break;

	  /* backslash - if escaping a quote within a quote  */
	case '\\':
	  if ((quote) && (*(instr + 1) == quote))
	    {
	      memmove (instr, instr + 1, strlen (instr));
  	      len--;
	    }
	   /* otherwise, its just a normal character */
	   else {
		if ( state == WHITESPACE ) {
			state = TOKENSTART;
		}
		}
	  break;


	  /* whitepsace */
	case ' ':
	case '\t':
	case '\r':
	case '\n':
	  if ((state == WORDSPACE) && (quote == '\0'))
	    {
	      state = WHITESPACE;
	      *instr = '\0';
	    }
	  break;

	case '\0':
	  break;
	  
	default: 
		if ( state == WHITESPACE ) {
			state = TOKENSTART;
		}

	}			/* end switch */
	  
	  if (state == TOKENSTART)
	  {
		  arg_count++;
	  if (arg_count > argc_slots)
	    {
	      argc_slots += ALLOC_CHUNK;
	      argv_array = (char **) realloc (argv_array, sizeof (char *) *
					      (argc_slots + ALLOC_CHUNK));
	    }

	  if (argv_array == (char) NULL)
	    {
	      return (-1);
	    }
	  argv_array[arg_count - 1] = instr;
	  state = WORDSPACE;
	}

      instr++;
      pos++;
    }

  argv_array[arg_count] = NULL;
  *argv = argv_array;
  return (arg_count);
}


/* uppercase an entire string, using toupper */
void
uppercase (char *instr)
{
  while (*instr != '\0')
    {
      *instr = toupper (*instr);
      instr++;
    }
}


/* lowercase an entire string, using tolower */
void
lowercase (char *instr)
{
  while (*instr != '\0')
    {
      *instr = tolower (*instr);
      instr++;
    }
}





#ifdef TEST_FRAMEWORK

main ()
{

  int argc, count;
  char **argv;
  char string[2000];

  strcpy (string,
	  "\\This\\ string will be  '' \"separated into\"  \"'\\\"'\" ' 15 ' elements.\n" \
	  "' including a multi-line\n"
	  "element' with a comment.  # This should not be parsed\n"
	  ";Nor should this\n"
	  "The End.");

  printf ("%s\n", string); 
  argc = argc_argv (string, &argv, "#;");
  for (count = 0; count < argc; count++)
    {
      printf ("%03d: [%s]\n", count, argv[count], "");
    }
  if (argc != 15 ) { puts ("Test FAILED"); } else { puts ("Test PASSED"); }
  free (argv);
}

#endif
